/*
Copyright (C) 2001 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
funcpars@hendriklipka.de
*/

package de.hendriklipka.funcpars;

import java.util.*;
import java.math.*;

public class FunctionParser
{
    public static final int ERROR_FORMEVAL=1;

    protected String _expression=null;
    protected Vector _RPNExpression=new Vector();
    protected Vector _tokenExpression=new Vector();
    protected Stack tokenStack=new Stack();
    protected int _error=0;

    /**
     *
    */
    public FunctionParser(String function)
    {
        _expression=function;
        tokenize();
        convertToRPN();
    }

    /**
     *
    */
    public double parse()
    {
        return parse(0,0,0);
    }

    /**
     *
    */
    public double parse(double x)
    {
        return parse(x,0,0);
    }

    /**
     *
    */
    public double parse(double x, double y)
    {
        return parse(x,y,0);
    }

    /**
     * @todo it would be faster if we have Double instead of double in the RPNField
    */
    public double parse(double x, double y, double z)
    {
        Stack stack=new Stack();
        for (int i=0;i<_RPNExpression.size();i++)
        {
            RPNField rf=(RPNField)_RPNExpression.elementAt(i);
            if (RPNField.TOK_NUMBER==rf._type)
            {
                stack.push(new Double(rf._doubleValue)) ;
            }
            else if (RPNField.TOK_PARAMETER==rf._type)
            {
                if (rf._strValue.equalsIgnoreCase("X"))
                {
                    stack.push(new Double(x)) ;
                }
                else if (rf._strValue.equalsIgnoreCase("Y"))
                {
                    stack.push(new Double(y)) ;
                }
                else if (rf._strValue.equalsIgnoreCase("Z"))
                {
                    stack.push(new Double(z)) ;
                }
            }
            else if (RPNField.TOK_OPERATOR==rf._type)
            {
                double h=rf._func.calculate(stack);
                //System.out.println(h);
                stack.push(new Double(h));
            } ;
        } ;
        double h=((Double)stack.pop()).doubleValue();
        return h;
    }

    /**
     *
    */
    protected RPNField getNextMorphem()
    {
        int pos=0;
        String chars="+-*/^";

        RPNField rf=new RPNField();

        char c=_expression.charAt(pos);
        if (Character.isDigit(c)) // numbers
        {
            while (true)
            {
                pos++;
                if (pos>=_expression.length())
                    break;
                if (!Character.isDigit(_expression.charAt(pos)))
                    break;
            }
            rf._strValue=_expression.substring(0,pos);
            rf._type=rf.TOK_NUMBER;
            rf._doubleValue=Double.parseDouble(rf._strValue);
            pos--;
        }
        else if (Character.isLetter(c))
        {
            while (true)
            {
                pos++;
                if (pos>=_expression.length())
                    break;
                if (!Character.isLetter(_expression.charAt(pos)))
                    break;
            }
            String s=_expression.substring(0,pos);
            rf._strValue=s;
            if (s.equalsIgnoreCase("X")||s.equalsIgnoreCase("Y")||s.equalsIgnoreCase("Z"))
            {
                rf._type=rf.TOK_PARAMETER;
            }
            else
            {
                rf._type=rf.TOK_OPERATOR;
            }
            pos--;
        }
        else if ('('==c)
        {
            rf._strValue="(";
            rf._type=rf.TOK_OPEN_BRACKET;
        }
        else if (')'==c)
        {
            rf._strValue=")";
            rf._type=rf.TOK_CLOSE_BRACKET;
        }
        else if (-1!=chars.indexOf(c))
        {
            rf._strValue=String.valueOf(c);
            rf._type=rf.TOK_OPERATOR;
        }
        else
        {
            System.err.println("error - unknown token!");
        }
        // determine funtion object, if funtion
        if (rf._type==rf.TOK_OPERATOR)
        {
            rf._func=determineFunctionForToken(rf._strValue);
        }
//        System.out.println("token="+rf._strValue);
        _expression=_expression.substring(pos+1);
        return rf;
    }

    /**
     */
    protected Function determineFunctionForToken(String token)
    {
        if (null==token)
            return null;
        for (int i=0;i<_functions.length;i++)
        {
            if (token.equalsIgnoreCase(_functions[i].getName()))
                return _functions[i];
        }
        return null;
    }

    /**
     *
    */
    protected void tokenize()
    {
        boolean type=true ;
        RPNField rf=null ;
        do
        {
            rf=getNextMorphem() ;
            if (null==rf)
            {
                _error=ERROR_FORMEVAL ;
                System.err.println("error: tokenize");
                break ;
            } ;
            if (RPNField.TOK_NUMBER==rf._type)
            {
                _tokenExpression.addElement(rf) ;
                type=false ;
            }
            else if (RPNField.TOK_PARAMETER==rf._type)
            {
                _tokenExpression.addElement(rf) ;
                type=false ;
            }
            else if (RPNField.TOK_OPEN_BRACKET==rf._type)
            {
                _tokenExpression.addElement(rf) ;
                type=true ;
            }
            else if (RPNField.TOK_CLOSE_BRACKET==rf._type)
            {
                _tokenExpression.addElement(rf) ;
                type=false ;
            }
            else if (RPNField.TOK_OPERATOR==rf._type)
            {
                if (type&&(rf._strValue.equals("-")))
                {
                    //System.out.println("negation!");
                    rf._strValue="~";
                    rf._func=determineFunctionForToken("~");
                }
                _tokenExpression.addElement(rf) ;
                type=true;
            }
            else if (RPNField.TOK_UNUSED==rf._type)
            {
                _error=ERROR_FORMEVAL ;
                System.err.println("error: tokenize");
                break ;
            } ;
        } while (_expression.length()>0) ;
    }

    /**
    */
    protected void convertToRPN()
    {
        int aktPos;
        int ii=_tokenExpression.size();
        for (aktPos=0;aktPos<ii;aktPos++)
        {
            RPNField h =(RPNField)_tokenExpression.elementAt(aktPos) ;
            RPNField hh=null;
            if (!tokenStack.empty())
            {
                hh=(RPNField)tokenStack.peek() ;
            } ;
            if (null!=h)
            {
                //System.out.println("h="+h._strValue);
            }
            switch (h._type)
            {
                case RPNField.TOK_OPERATOR :
                                        if (null!=hh)
                                        {
                                            //System.out.println("hh="+hh._strValue);
                                        }
                                        if (tokenStack.empty())
                                        {
                                            //System.out.println("op1 - push h to stack");
                                            tokenStack.push(h) ;
                                        }
                                        else if ((RPNField.TOK_OPEN_BRACKET==hh._type)||(getFuncNum(hh._func)>getFuncNum(h._func)))
                                        {
                                            //System.out.println("op2 - push h to stack");
                                            tokenStack.push(h) ;
                                        }
                                        else
                                        {
                                            //System.out.println("op3");
                                            while (!tokenStack.empty()&&(RPNField.TOK_OPEN_BRACKET!=hh._type)&&(getFuncNum(hh._func)<getFuncNum(h._func)))
                                            {
                                                //System.out.println("adding to RPN: "+((RPNField)tokenStack.peek())._strValue);
                                                _RPNExpression.addElement(tokenStack.pop()) ;
                                                if (tokenStack.empty())
                                                    break ;
                                                hh=(RPNField)tokenStack.peek() ;
                                            }
                                            //System.out.println("pushing h to stack");
                                            tokenStack.push(h) ;
                                        } ;
                                        break ;
                case RPNField.TOK_NUMBER:
                                        //System.out.println("num - adding to RPN");
                                        _RPNExpression.addElement(h) ;
                                        break ;
                case RPNField.TOK_PARAMETER:
                                        //System.out.println("param - adding to RPN");
                                        _RPNExpression.addElement(h) ;
                                        break ;
                case RPNField.TOK_OPEN_BRACKET:
                                        //System.out.println("open - push to stack");
                                        tokenStack.push(h) ;
                                        break ;
                case RPNField.TOK_CLOSE_BRACKET:
                                        //System.out.println("found closing bracket");
                                        while (((RPNField)tokenStack.peek())._type!=RPNField.TOK_OPEN_BRACKET)
                                        {
                                            //System.out.println("adding to RPN"+((RPNField)tokenStack.peek())._strValue);
                                            _RPNExpression.addElement(tokenStack.pop()) ;
                                            //System.out.println("next="+((RPNField)tokenStack.peek())._strValue);
                                        }
                                        //System.out.println("popping (");
                                        tokenStack.pop() ; // '(' rauswerfen
                                        break ;
            }
        }
        while (!tokenStack.empty())
        {
            _RPNExpression.addElement(tokenStack.pop()) ;
        }
    }

    /**
    */
    public void DumpRPN()
    {
        for (int i=0;i<_RPNExpression.size();i++)
        {
            RPNField rf=(RPNField)_RPNExpression.elementAt(i);
            System.out.println("#"+i+"="+rf._strValue);
        }
    }

    /**
     * @todo implement
    */
    public void DumpToken()
    {
    }

    /**
    */
    protected int getFuncNum(Function func)
    {
        if (null==func)
            return -1;
        for (int i=0;i<_functions.length;i++)
        {
            if (_functions[i]==func)
                return i;
        }
        return -1;
    }

    /**
     *
    */
    protected static abstract class Function
    {
        public abstract double calculate(Stack stack);
        public abstract String getName();
    }

    /**
     * all supported functions
     * NOTE: order is significant!
    */
    protected Function[] _functions={
        new Function(){public String getName(){return "sin";} public double calculate(Stack stack) {double d1=((Double)stack.pop()).doubleValue();return Math.sin(d1);}},
        new Function(){public String getName(){return "~";} public double calculate(Stack stack) {double d1=((Double)stack.pop()).doubleValue();return -d1;}},
        new Function(){public String getName(){return "*";} public double calculate(Stack stack) {double d1=((Double)stack.pop()).doubleValue();double d2=((Double)stack.pop()).doubleValue();return d1*d2;}},
        new Function(){public String getName(){return "/";} public double calculate(Stack stack) {double d1=((Double)stack.pop()).doubleValue();double d2=((Double)stack.pop()).doubleValue();return d2/d1;}},
        new Function(){public String getName(){return "^";} public double calculate(Stack stack) {double d1=((Double)stack.pop()).doubleValue();double d2=((Double)stack.pop()).doubleValue();return Math.pow(d1,d2);}},
        new Function(){public String getName(){return "+";} public double calculate(Stack stack) {double d1=((Double)stack.pop()).doubleValue();double d2=((Double)stack.pop()).doubleValue();return d1+d2;}},
        new Function(){public String getName(){return "-";} public double calculate(Stack stack) {double d1=((Double)stack.pop()).doubleValue();double d2=((Double)stack.pop()).doubleValue();return d2-d1;}}
    };

    /**
     *
    */
    protected class RPNField
    {
        public static final int TOK_UNUSED=0;
        public static final int TOK_NUMBER=1;
        public static final int TOK_OPERATOR=2;
        public static final int TOK_OPEN_BRACKET=3;
        public static final int TOK_CLOSE_BRACKET=4;
        public static final int TOK_PARAMETER=5;

        public int _type=TOK_UNUSED ; // Type of this element
        public Function _func=null; // function for this RPN element
        public String _strValue=null;
        public double _doubleValue ; // numeric value
    }
}